use crate::model::SectionRef;
use crate::option_future::OptionFuture;
use crate::section_interface::SectionInterface;
use std::{
    mem::swap,
    sync::{
        atomic::{AtomicI32, Ordering},
        Arc,
    },
    time::Duration,
};
use thiserror::Error;
use tokio::{
    spawn,
    sync::{broadcast, mpsc, oneshot},
    time::{delay_for, Instant},
};
use tracing::{debug, trace, trace_span, warn};
use tracing_futures::Instrument;

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct SectionRunHandle(i32);

#[derive(Debug)]
struct SectionRunnerInner {
    next_run_id: AtomicI32,
}

impl SectionRunnerInner {
    fn new() -> Self {
        Self {
            next_run_id: AtomicI32::new(1),
        }
    }
}

#[derive(Debug)]
enum RunnerMsg {
    Quit(oneshot::Sender<()>),
    QueueRun(SectionRunHandle, SectionRef, Duration),
    CancelRun(SectionRunHandle),
    CancelAll,
    Pause,
    Unpause,
    Subscribe(oneshot::Sender<SectionEventRecv>),
}

#[derive(Clone, Debug)]
pub enum SectionEvent {
    RunStart(SectionRunHandle, SectionRef),
    RunFinish(SectionRunHandle, SectionRef),
    RunPause(SectionRunHandle, SectionRef),
    RunUnpause(SectionRunHandle, SectionRef),
    RunCancel(SectionRunHandle, SectionRef),
    RunnerPause,
    RunnerUnpause,
}

pub type SectionEventRecv = broadcast::Receiver<SectionEvent>;
type SectionEventSend = broadcast::Sender<SectionEvent>;

const EVENT_CAPACITY: usize = 8;

#[derive(Clone, Debug, PartialEq)]
pub enum SecRunState {
    Waiting,
    Running {
        start_time: Instant,
    },
    Finished,
    Cancelled,
    Paused {
        start_time: Instant,
        pause_time: Instant,
    },
}

#[derive(Clone, Debug)]
pub struct SecRun {
    handle: SectionRunHandle,
    section: SectionRef,
    duration: Duration,
    total_duration: Duration,
    state: SecRunState,
}

impl SecRun {
    fn new(handle: SectionRunHandle, section: SectionRef, duration: Duration) -> Self {
        Self {
            handle,
            section,
            duration,
            total_duration: duration,
            state: SecRunState::Waiting,
        }
    }

    pub fn is_running(&self) -> bool {
        matches!(self.state, SecRunState::Running{..})
    }

    #[allow(dead_code)]
    pub fn is_paused(&self) -> bool {
        matches!(self.state, SecRunState::Paused{..})
    }
}

pub type SecRunQueue = im::Vector<Arc<SecRun>>;

#[derive(Clone, Debug)]
pub struct SecRunnerState {
    run_queue: SecRunQueue,
    paused: bool,
}

impl Default for SecRunnerState {
    fn default() -> Self {
        Self {
            run_queue: SecRunQueue::default(),
            paused: false,
        }
    }
}

struct RunnerTask {
    interface: Arc<dyn SectionInterface + Sync>,
    msg_recv: mpsc::Receiver<RunnerMsg>,
    running: bool,
    delay_future: OptionFuture<tokio::time::Delay>,
    event_send: Option<SectionEventSend>,
    quit_tx: Option<oneshot::Sender<()>>,
}

impl RunnerTask {
    fn new(
        interface: Arc<dyn SectionInterface + Sync>,
        msg_recv: mpsc::Receiver<RunnerMsg>,
    ) -> Self {
        Self {
            interface,
            msg_recv,
            running: true,
            delay_future: None.into(),
            event_send: None,
            quit_tx: None,
        }
    }

    fn send_event(&mut self, event: SectionEvent) {
        if let Some(event_send) = &mut self.event_send {
            match event_send.send(event) {
                Ok(_) => {}
                Err(_closed) => {
                    self.event_send = None;
                }
            }
        }
    }

    fn subscribe_event(&mut self) -> SectionEventRecv {
        match &mut self.event_send {
            Some(event_send) => event_send.subscribe(),
            None => {
                let (event_send, event_recv) = broadcast::channel(EVENT_CAPACITY);
                self.event_send = Some(event_send);
                event_recv
            }
        }
    }

    fn start_run(&mut self, run: &mut Arc<SecRun>) {
        use SecRunState::*;
        let run = Arc::make_mut(run);
        debug!(section_id = run.section.id, "starting running section");
        self.interface
            .set_section_state(run.section.interface_id, true);
        run.state = Running {
            start_time: Instant::now(),
        };
        self.send_event(SectionEvent::RunStart(
            run.handle.clone(),
            run.section.clone(),
        ));
    }

    fn finish_run(&mut self, run: &mut Arc<SecRun>) {
        let run = Arc::make_mut(run);
        if run.is_running() {
            debug!(section_id = run.section.id, "finished running section");
            self.interface
                .set_section_state(run.section.interface_id, false);
            run.state = SecRunState::Finished;
            self.send_event(SectionEvent::RunFinish(
                run.handle.clone(),
                run.section.clone(),
            ));
        } else {
            warn!(
                section_id = run.section.id,
                state = debug(&run.state),
                "cannot finish run which is not running"
            );
        }
    }

    fn cancel_run(&mut self, run: &mut Arc<SecRun>) {
        let run = Arc::make_mut(run);
        if run.is_running() {
            debug!(section_id = run.section.id, "cancelling running section");
            self.interface
                .set_section_state(run.section.interface_id, false);
        }
        run.state = SecRunState::Cancelled;
        self.send_event(SectionEvent::RunCancel(
            run.handle.clone(),
            run.section.clone(),
        ));
    }

    fn pause_run(&mut self, run: &mut Arc<SecRun>) {
        use SecRunState::*;
        let run = Arc::make_mut(run);
        let new_state = match run.state {
            Running { start_time } => {
                debug!(section_id = run.section.id, "pausing running section");
                self.interface
                    .set_section_state(run.section.interface_id, false);
                Paused {
                    start_time,
                    pause_time: Instant::now(),
                }
            }
            Waiting => {
                debug!(section_id = run.section.id, "pausing waiting section");
                Paused {
                    start_time: Instant::now(),
                    pause_time: Instant::now(),
                }
            }
            Finished | Cancelled | Paused { .. } => {
                return;
            }
        };
        run.state = new_state;
        self.send_event(SectionEvent::RunPause(
            run.handle.clone(),
            run.section.clone(),
        ));
    }

    fn unpause_run(&mut self, run: &mut Arc<SecRun>) {
        use SecRunState::*;
        let run = Arc::make_mut(run);
        match run.state {
            Paused {
                start_time,
                pause_time,
            } => {
                debug!(section_id = run.section.id, "unpausing section");
                self.interface
                    .set_section_state(run.section.interface_id, true);
                run.state = Running {
                    start_time: Instant::now(),
                };
                let ran_for = pause_time - start_time;
                run.duration -= ran_for;
                self.send_event(SectionEvent::RunUnpause(
                    run.handle.clone(),
                    run.section.clone(),
                ));
            }
            Waiting | Finished | Cancelled | Running { .. } => {
                warn!(
                    section_id = run.section.id,
                    state = debug(&run.state),
                    "can only unpause paused section"
                );
            }
        }
    }

    fn process_queue(&mut self, state: &mut SecRunnerState) {
        use SecRunState::*;
        while let Some(current_run) = state.run_queue.front_mut() {
            let run_finished = match (&current_run.state, state.paused) {
                (Waiting, false) => {
                    self.start_run(current_run);
                    self.delay_future = Some(delay_for(current_run.duration)).into();
                    false
                }
                (Running { start_time }, false) => {
                    let time_to_finish = start_time.elapsed() >= current_run.duration;
                    if time_to_finish {
                        self.finish_run(current_run);
                        self.delay_future = None.into();
                    }
                    time_to_finish
                }
                (Paused { .. }, false) => {
                    self.unpause_run(current_run);
                    self.delay_future = Some(delay_for(current_run.duration)).into();
                    false
                }
                (Waiting, true) => {
                    self.pause_run(current_run);
                    false
                }
                (Running { .. }, true) => {
                    self.pause_run(current_run);
                    self.delay_future = None.into();
                    false
                }
                (Paused { .. }, true) => false,
                (Cancelled, _) | (Finished, _) => true,
            };

            if run_finished {
                state.run_queue.pop_front();
            } else {
                break;
            }
        }
    }

    fn handle_msg(&mut self, msg: Option<RunnerMsg>, state: &mut SecRunnerState) {
        let msg = msg.expect("SectionRunner channel closed");
        use RunnerMsg::*;
        trace!(msg = debug(&msg), "runner_task recv");
        match msg {
            Quit(quit_tx) => {
                self.quit_tx = Some(quit_tx);
                self.running = false;
            }
            QueueRun(handle, section, duration) => {
                state
                    .run_queue
                    .push_back(Arc::new(SecRun::new(handle, section, duration)));
            }
            CancelRun(handle) => {
                for run in state.run_queue.iter_mut() {
                    if run.handle != handle {
                        continue;
                    }
                    trace!(handle = handle.0, "cancelling run by handle");
                    self.cancel_run(run);
                }
            }
            CancelAll => {
                let mut old_runs = SecRunQueue::new();
                swap(&mut old_runs, &mut state.run_queue);
                trace!(count = old_runs.len(), "cancelling all runs");
                for mut run in old_runs {
                    self.cancel_run(&mut run);
                }
            }
            Pause => {
                state.paused = true;
                self.send_event(SectionEvent::RunnerPause);
            }
            Unpause => {
                state.paused = false;
                self.send_event(SectionEvent::RunnerUnpause);
            }
            Subscribe(res_send) => {
                let event_recv = self.subscribe_event();
                // Ignore error if channel closed
                let _ = res_send.send(event_recv);
            }
        }
    }

    async fn run_impl(mut self) {
        let mut state = SecRunnerState::default();

        while self.running {
            self.process_queue(&mut state);
            let delay_done = || {
                trace!("delay done");
            };
            tokio::select! {
                msg = self.msg_recv.recv() => self.handle_msg(msg, &mut state),
                _ = &mut self.delay_future, if self.delay_future.is_some() => delay_done()
            };
        }

        if let Some(quit_tx) = self.quit_tx.take() {
            let _ = quit_tx.send(());
        }
    }

    async fn run(self) {
        let span = trace_span!("section_runner task");

        self.run_impl().instrument(span).await;
    }
}

#[derive(Debug, Clone, Error)]
#[error("the SectionRunner channel is closed")]
pub struct ChannelClosed;

pub type Result<T, E = ChannelClosed> = std::result::Result<T, E>;

impl<T> From<mpsc::error::SendError<T>> for ChannelClosed {
    fn from(_: mpsc::error::SendError<T>) -> Self {
        Self
    }
}

impl From<oneshot::error::RecvError> for ChannelClosed {
    fn from(_: oneshot::error::RecvError) -> Self {
        Self
    }
}

#[derive(Clone, Debug)]
pub struct SectionRunner {
    inner: Arc<SectionRunnerInner>,
    msg_send: mpsc::Sender<RunnerMsg>,
}

#[allow(dead_code)]
impl SectionRunner {
    pub fn new(interface: Arc<dyn SectionInterface + Sync>) -> Self {
        let (msg_send, msg_recv) = mpsc::channel(8);
        spawn(RunnerTask::new(interface, msg_recv).run());
        Self {
            inner: Arc::new(SectionRunnerInner::new()),
            msg_send,
        }
    }

    pub async fn quit(&mut self) -> Result<()> {
        let (quit_tx, quit_rx) = oneshot::channel();
        self.msg_send.send(RunnerMsg::Quit(quit_tx)).await?;
        quit_rx.await?;
        Ok(())
    }

    pub async fn queue_run(
        &mut self,
        section: SectionRef,
        duration: Duration,
    ) -> Result<SectionRunHandle> {
        let run_id = self.inner.next_run_id.fetch_add(1, Ordering::Relaxed);
        let handle = SectionRunHandle(run_id);
        self.msg_send
            .send(RunnerMsg::QueueRun(handle.clone(), section, duration))
            .await?;
        Ok(handle)
    }

    pub async fn cancel_run(&mut self, handle: SectionRunHandle) -> Result<()> {
        self.msg_send.send(RunnerMsg::CancelRun(handle)).await?;
        Ok(())
    }

    pub async fn cancel_all(&mut self) -> Result<()> {
        self.msg_send.send(RunnerMsg::CancelAll).await?;
        Ok(())
    }

    pub async fn pause(&mut self) -> Result<()> {
        self.msg_send.send(RunnerMsg::Pause).await?;
        Ok(())
    }

    pub async fn unpause(&mut self) -> Result<()> {
        self.msg_send.send(RunnerMsg::Unpause).await?;
        Ok(())
    }

    pub async fn subscribe(&mut self) -> Result<SectionEventRecv> {
        let (res_send, res_recv) = oneshot::channel();
        self.msg_send.send(RunnerMsg::Subscribe(res_send)).await?;
        let event_recv = res_recv.await?;
        Ok(event_recv)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::section_interface::MockSectionInterface;
    use crate::{
        model::{Section, Sections},
        trace_listeners::{EventListener, Filters, SpanFilters, SpanListener},
    };
    use assert_matches::assert_matches;
    use im::ordmap;
    use tracing_subscriber::prelude::*;

    #[tokio::test]
    async fn test_quit() {
        let quit_msg = EventListener::new(
            Filters::new()
                .target("sprinklers_rs::section_runner")
                .message("runner_task recv")
                .field_value("msg", "Quit"),
        );
        let task_span = SpanListener::new(
            SpanFilters::new()
                .target("sprinklers_rs::section_runner")
                .name("section_runner task"),
        );
        let subscriber = tracing_subscriber::registry()
            .with(quit_msg.clone())
            .with(task_span.clone());
        let _sub = tracing::subscriber::set_default(subscriber);

        let interface = MockSectionInterface::new(6);
        let mut runner = SectionRunner::new(Arc::new(interface));
        tokio::task::yield_now().await;
        runner.quit().await.unwrap();

        assert_eq!(quit_msg.get_count(), 1);
        assert!(task_span.get_exit_count() > 1);
    }

    fn make_sections_and_interface() -> (Sections, Arc<MockSectionInterface>) {
        let interface = Arc::new(MockSectionInterface::new(2));
        let sections: Sections = ordmap![
            1 => Section {
                id: 1,
                name: "Section 1".into(),
                interface_id: 0,
            }.into(),
            2 => Section {
                id: 2,
                name: "Section 2".into(),
                interface_id: 1,
            }.into()
        ];
        (sections, interface)
    }

    fn assert_section_states(interface: &MockSectionInterface, states: &[bool]) {
        for (id, state) in states.iter().enumerate() {
            assert_eq!(
                interface.get_section_state(id as u32),
                *state,
                "section interface id {} did not match",
                id
            );
        }
    }

    async fn advance(dur: Duration) {
        // HACK: advance should really be enough, but we need another yield_now
        tokio::time::pause();
        tokio::time::advance(dur).await;
        tokio::task::yield_now().await;
        tokio::time::resume();
    }

    #[tokio::test]
    async fn test_queue() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        assert_section_states(&interface, &[false, false]);

        // Queue single section, make sure it runs
        runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_section_states(&interface, &[true, false]);

        advance(Duration::from_secs(11)).await;

        assert_section_states(&interface, &[false, false]);

        // Queue two sections, make sure they run one at a time
        runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(11)).await;

        assert_section_states(&interface, &[true, false]);

        advance(Duration::from_secs(10)).await;

        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[tokio::test]
    async fn test_cancel_run() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        let run1 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let _run2 = runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run3 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_section_states(&interface, &[false, true]);

        runner.cancel_run(run1).await.unwrap();
        tokio::task::yield_now().await;

        assert_section_states(&interface, &[true, false]);

        runner.cancel_run(run3).await.unwrap();
        advance(Duration::from_secs(11)).await;

        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[tokio::test]
    async fn test_cancel_all() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[tokio::test]
    async fn test_pause() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        let _run1 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run2 = runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let _run3 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        runner.pause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        advance(Duration::from_secs(10)).await;
        assert_section_states(&interface, &[false, false]);

        runner.unpause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(8)).await;
        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(2)).await;
        assert_section_states(&interface, &[true, false]);

        runner.pause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        // cancel paused run
        runner.cancel_run(run2).await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        runner.unpause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(11)).await;
        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[tokio::test]
    async fn test_event() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        let mut event_recv = runner.subscribe().await.unwrap();

        let run1 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run2 = runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run3 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        assert_matches!(
            event_recv.recv().await,
            Ok(SectionEvent::RunStart(handle, _))
            if handle == run1
        );

        runner.pause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunnerPause));
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunPause(handle, _)) if handle == run1);

        runner.unpause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunnerUnpause));
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunUnpause(handle, _)) if handle == run1);

        advance(Duration::from_secs(11)).await;
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunFinish(handle, _)) if handle == run1);
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunStart(handle, _)) if handle == run2);

        runner.pause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunnerPause));
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunPause(handle, _)) if handle == run2);

        // cancel paused run
        runner.cancel_run(run2.clone()).await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunCancel(handle, _)) if handle == run2);
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunPause(handle, _)) if handle == run3);

        runner.unpause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunnerUnpause));
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunUnpause(handle, _)) if handle == run3);

        advance(Duration::from_secs(11)).await;
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunFinish(handle, _)) if handle == run3);

        runner.quit().await.unwrap();
    }
}
