use crate::{
    mqtt_interface::MqttInterface,
    program_runner::{ProgramEvent, ProgramEventRecv},
    section_runner::{SectionEvent, SectionEventRecv},
};
use tokio::{
    select,
    sync::{broadcast, oneshot},
    task::JoinHandle,
};
use tracing::{trace, trace_span, warn};
use tracing_futures::Instrument;

struct UpdateListenerTask {
    section_events: SectionEventRecv,
    program_events: ProgramEventRecv,
    mqtt_interface: MqttInterface,
    running: bool,
}

impl UpdateListenerTask {
    async fn handle_section_event(
        &mut self,
        event: Result<SectionEvent, broadcast::RecvError>,
    ) -> eyre::Result<()> {
        let event = match event {
            Ok(ev) => ev,
            Err(broadcast::RecvError::Closed) => {
                trace!("section events channel closed");
                self.running = false;
                return Ok(());
            }
            Err(broadcast::RecvError::Lagged(n)) => {
                warn!("section events lagged by {}", n);
                return Ok(());
            }
        };
        if let Some((sec_id, state)) = match event {
            SectionEvent::RunStart(_, sec) | SectionEvent::RunUnpause(_, sec) => {
                Some((sec.id, true))
            }
            SectionEvent::RunFinish(_, sec)
            | SectionEvent::RunPause(_, sec)
            | SectionEvent::RunCancel(_, sec) => Some((sec.id, false)),
            SectionEvent::RunnerPause | SectionEvent::RunnerUnpause => None,
        } {
            self.mqtt_interface
                .publish_section_state(sec_id, state)
                .await?;
        }
        Ok(())
    }

    async fn handle_program_event(
        &mut self,
        event: Result<ProgramEvent, broadcast::RecvError>,
    ) -> eyre::Result<()> {
        let event = match event {
            Ok(ev) => ev,
            Err(broadcast::RecvError::Closed) => {
                trace!("section events channel closed");
                self.running = false;
                return Ok(());
            }
            Err(broadcast::RecvError::Lagged(n)) => {
                warn!("section events lagged by {}", n);
                return Ok(());
            }
        };
        let (program_id, running) = match event {
            ProgramEvent::RunStart(prog) => (prog.id, true),
            ProgramEvent::RunFinish(prog) | ProgramEvent::RunCancel(prog) => (prog.id, false),
        };
        self.mqtt_interface
            .publish_program_running(program_id, running)
            .await?;
        Ok(())
    }

    async fn run_impl(&mut self) -> eyre::Result<()> {
        while self.running {
            select! {
                section_event = self.section_events.recv() => {
                    self.handle_section_event(section_event).await?
                }
                program_event = self.program_events.recv() => {
                    self.handle_program_event(program_event).await?
                }
            };
        }
        Ok(())
    }

    async fn run_or_quit(mut self, mut quit_rx: oneshot::Receiver<()>) -> eyre::Result<()> {
        select! {
            _ = &mut quit_rx => {
                self.running = false;
                Ok(())
            }
            res = self.run_impl() => {
                res
            }
        }
    }

    async fn run(self, quit_rx: oneshot::Receiver<()>) {
        let span = trace_span!("update_listener task");

        self.run_or_quit(quit_rx)
            .instrument(span)
            .await
            .expect("error in UpdateListenerTask");
    }
}

pub struct UpdateListener {
    quit_tx: oneshot::Sender<()>,
    join_handle: JoinHandle<()>,
}

impl UpdateListener {
    pub fn start(
        section_events: SectionEventRecv,
        program_events: ProgramEventRecv,
        mqtt_interface: MqttInterface,
    ) -> Self {
        let task = UpdateListenerTask {
            section_events,
            program_events,
            mqtt_interface,
            running: true,
        };
        let (quit_tx, quit_rx) = oneshot::channel();
        let join_handle = tokio::spawn(task.run(quit_rx));
        Self {
            quit_tx,
            join_handle,
        }
    }

    pub async fn quit(self) -> eyre::Result<()> {
        let _ = self.quit_tx.send(());
        self.join_handle.await?;
        Ok(())
    }
}
